/*
  vt1xx terminal emulation by
  Rene Rebe <rene@exactcode.de>
  
  Lossy based on code of bogl-term by:
  Edmund GRIMLEY EVANS <edmundo@rano.org>
  
  Additional information from the Linux kernel driver/char/vt.c by:
  Linus Torvalds
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA. */

#define _XOPEN_SOURCE 500
#include <errno.h>
#include <fcntl.h>
#include <locale.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>

#include "eTerm.hh"

#include "EvasHelper.hh"
#include "X11Helper.hh"

#define XK_MISCELLANY
#define XK_XKB_KEYS
#include <X11/keysymdef.h>

#include <iostream>

static int child_pid = 0;
static struct termios ttysave;

#include "TermUtil.hh"

static int quit = 0;

void sigterm (int sig)
{
  quit = 1;
}

// screen width / height
const int sw = 80;
const int sh = 25;

int main (int argc, char *argv[])
{
  int ret;
  char buf[8192];
  struct timeval tv;
  int ptyfd, ttyfd;
  char *locale = "", *command = NULL;
  int i;
  char o = ' ';
  
  int win_w = 1;
  int win_h = 1;
  
  // evas
  EvasCanvas* evas;
  eTerm* term;
  
  // TODO: move to the X11Helper ...
  
  XSetWindowAttributes attr;
  XClassHint           chint;
  
  // X11 stuff
  Display* dpy;
  int scr;
  Visual* visual;
  Window  win;
  int depth;
  
  dpy = XOpenDisplay (NULL);
  if (!dpy) {
    return false;
  }

  scr = DefaultScreen (dpy);

  Window root = RootWindow (dpy, scr);

  attr.backing_store = NotUseful;
  attr.border_pixel = 0;
  attr.background_pixmap = None;
  attr.event_mask =
    ExposureMask |
    ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
    KeyPressMask | KeyReleaseMask |
    StructureNotifyMask;
  attr.bit_gravity = ForgetGravity;
  /*  attr.override_redirect = 1; */

  visual = DefaultVisual (dpy, scr);
  depth = DefaultDepth(dpy, scr);
  attr.colormap = DefaultColormap(dpy, scr);
  
  win = XCreateWindow (dpy, root,
		       (X11Window::Width(dpy, 0) - win_w) / 2,
		       (X11Window::Height(dpy, 0) - win_h) / 2,
		       win_w, win_h,
		       0,
		       depth,
		       InputOutput,
		       visual,
		       /*CWOverrideRedirect | */ 
		       CWBackingStore | CWColormap |
		       CWBackPixmap | CWBorderPixel |
		       CWBitGravity | CWEventMask, &attr);

  XStoreName (dpy, win, "escreen");
  chint.res_name = "escreen";
  chint.res_class = "Main";
  XSetClassHint (dpy, win, &chint);
  
  XSync (dpy, False);

  evas = new EvasCanvas ();
  evas->OutputMethod ("software_x11");
  
  evas->OutputSize (win_w, win_h);
  evas->OutputViewport (0, 0, win_w, win_h);
  
  Evas_Engine_Info_Software_X11* einfo;
  
  einfo = (Evas_Engine_Info_Software_X11*) evas->EngineInfo ();
  
  /* the following is specific to the engine */
  einfo->info.display = dpy;
  einfo->info.visual = visual;
  einfo->info.colormap = attr.colormap;
  einfo->info.drawable = win;
  einfo->info.depth = depth;
  einfo->info.rotation = 0;
  einfo->info.debug = 0;
  
  evas->EngineInfo ( (Evas_Engine_Info*)einfo );
  
  /* Setup */
  evas->FontPathPrepend ("/usr/X11/lib/X11/fonts/TTF/");
  evas->FontPathPrepend ("/usr/X11/lib/X11/fonts/TrueType/");
  evas->FontPathPrepend ("/opt/e17/share/evas/data/");
  
  if (true) {
    evas->ImageCache (1024 * 1024);
    evas->FontCache (256 * 1024);
  }
  else {
    evas->ImageCache (0);
    evas->FontCache (0);
  }
  
  term = new eTerm (evas, sw, sh);
  
  win_w = term->Width();
  win_h = term->Height();
  
  evas->OutputSize (win_w, win_h);
  evas->OutputViewport (0, 0, win_w, win_h);
  
  X11Window::Resize (dpy, win, win_w, win_h);
  XMapWindow (dpy, win);

  for (i = 1 ; i < argc ; ++i)
    if (argv[i][0] == '-')
      switch (argv[i][1])
	{
	case 'f':
	case 'l':
	  o = argv[i][1];
	  break;

	default:
	  printf ("unknown option: %c\n", argv[i][1]);
	}
    else
      switch (o)
	{
	case ' ':
	  command = argv[i];
	  break;

	case 'l':
	  locale = argv[i];
	  o = ' ';
	  break;
	}

  setlocale (LC_CTYPE, locale);
  
  // TODO: avoid this?
  tcgetattr (0, &ttysave);
  //ttysave
  //  tio->c_cc[VEOF] = CEOF;
  //  tio->c_cc[VEOL] = VDISABLE;
  //ttysave.c_cflag |= ICANON | ECHO;
  
  if (get_ptytty (&ptyfd, &ttyfd)) {
    perror("can't get a pty");
    exit(1);
  }
  
  spawn_shell (ptyfd, ttyfd, command == NULL ? "/bin/sh" : command);
  
  evas->Render ();
  
  // signal(SIGHUP, reload_font);
  signal (SIGTERM, sigterm);

  set_window_size (ttyfd, sw, sh);

  for (;;) {
    fd_set fds;
    int max = 0;

    if (quit)
      break;
    
    // process X11 events ...
    // TODO: move into X11 Helper ...
    
    XEvent ev;
    while (XCheckMaskEvent (dpy,
			    ExposureMask |
			    StructureNotifyMask |
			    KeyPressMask |
			    KeyReleaseMask |
			    ButtonPressMask |
			    ButtonReleaseMask |
			    PointerMotionMask, &ev))
      {
	Evas_Button_Flags flags = EVAS_BUTTON_NONE;
	switch (ev.type)
	  {
	  case KeyPress:
	    {
	      KeySym ks;
	      XLookupString ((XKeyEvent*)&ev, 0, 0, &ks, NULL);
	      
	      bool handled = false;
	      
	      // handle Ctrl Alt
	      int state = ev.xkey.state & (ControlMask|Mod1Mask);
	      
	      // not both at the same time ...
	      if (state == (ControlMask | Mod1Mask)) {
		//std::cerr << "both" << std::endl;
		break;
	      }
	      
	      if (state == ControlMask) {
		//std::cerr << "control" << std::endl;
		// special sequence for the characters
		char c = tolower (ks);

		if (c >= 'a' && c <= 'z') {
		  c -= 'a' - 1;
		  write (ptyfd, &c, 1);
		  handled = true;
		}
		else if (c == ' ') {
		  c = '\0';
		  write (ptyfd, &c, 1);
		  handled = true;
		}
		
		if (!handled)
		  std::cerr << "Unhandled control key sequence." << std::endl;
		break;
	      }
	      else if (state == Mod1Mask) {
		//std::cerr << "alt" << std::endl;
		// mark alt
		write (ptyfd, "\33", 1);
		// + the symbol, so let it go ...
	      }
	      
	      // non Ctrl or Alt state, ...
	      {
		char c = ks;
		if (ks >= ' ' && ks < 127) {
		  write (ptyfd, &c, 1);
		  handled = true;
		}
		else {
		  const char* c = NULL;
		  switch (ks)
		    {
		    case XK_Shift_L:
		    case XK_Shift_R:
		    case XK_Control_L:
		    case XK_Control_R:
		    case XK_Caps_Lock:
		    case XK_Shift_Lock:

		    case XK_Meta_L:
		    case XK_Meta_R:
		    case XK_Alt_L:
		    case XK_Alt_R:
		    case XK_Super_L:
		    case XK_Super_R:
		    case XK_Hyper_L:
		    case XK_Hyper_R:
		      
		    case XK_ISO_Level3_Shift:

		      handled = true; // NOP
		      break;
		      
		    case XK_BackSpace:   c = "\177"; break;
		    case XK_Tab:         c = "\t"; break;
		    case XK_Return:      c = "\r"; break;
		    case XK_Escape:      c = "\33"; break;
		      
		    case XK_Up:          c = "\033[A"; break;
		    case XK_Down:        c = "\033[B"; break;
		    case XK_Right:       c = "\033[C"; break;
		    case XK_Left:        c = "\033[D"; break;
		    case XK_Home:        c = "\033[1~"; break;
		    case XK_Insert:      c = "\033[2~"; break;
		    case XK_Delete:      c = "\033[3~"; break;
		    case XK_End:         c = "\033[4~"; break;
		    case XK_Page_Up:     c = "\033[5~"; break;
		    case XK_Page_Down:   c = "\033[6~"; break;
		    
		      /* TODO: only F6++ do work ? */

		    case XK_F1:   c = "\033[[\7A"; break;
		    case XK_F2:   c = "\033[[\7B"; break;
		    case XK_F3:   c = "\033[[\7C"; break;
		    case XK_F4:   c = "\033[[\7D"; break;
		    case XK_F5:   c = "\033[[\7E"; break;
		    case XK_F6:   c = "\033[17~"; break;
		    case XK_F7:   c = "\033[18~"; break;
		    case XK_F8:   c = "\033[19~"; break;
		    case XK_F9:   c = "\033[20~"; break;
		    case XK_F10:  c = "\033[21~"; break;
		    case XK_F11:  c = "\033[23~"; break;
		    case XK_F12:  c = "\033[24~"; break;
		      
		      // case XK_Next: break;
		      
		      // case XK_Prior: break;

		      // case XK_Begin: break;
		    }
		  if (c) {
		    //std::cerr << "char" << std::endl;
		
		    handled = true;
		    write (ptyfd, c, strlen(c));
		  }
		}
		if (!handled)
		  {
		    std::cout << "unhandled key" << std::endl;
		    std::cout << "key: " << ev.xkey.keycode << std::endl;
		    std::cout << "state: " << ev.xkey.state << std::endl;
		    std::cout << "sym: " << ks << std::endl;
		  }
	      }
	    }
	    break;
	    
	    case Expose:
	    evas->DamageRectangleAdd (ev.xexpose.x,
				      ev.xexpose.y,
				      ev.xexpose.width,
				      ev.xexpose.height);
	    break;
	  case ConfigureNotify:
	    evas->OutputSize (ev.xconfigure.width,
			      ev.xconfigure.height);
	    evas->OutputViewport (0, 0,
				  ev.xconfigure.width,
				  ev.xconfigure.height);
	    break;
	  default:
	    break;
	  }
      }
    
    FD_ZERO(&fds);
    FD_SET(ptyfd, &fds);
    if (ptyfd > max)
      max = ptyfd;
    
    tv.tv_sec = 0;
    tv.tv_usec = 1000;
    ret = select (max+1, &fds, NULL, NULL, &tv);
    
    if (FD_ISSET(ptyfd,&fds)) {
      ret = read (ptyfd, buf, sizeof(buf));
      if (ret > 0) {
	// output to screen
	term->out (buf, ret);
      }
    }
    
    term->redraw ();
    evas->Render ();
    
    XFlush (dpy);
  }
  
  delete evas;
  evas = 0;  
  
  return 0;
}
